#include "common.h"

void putxmlstr(FILE *f, const char *str)
{
  int i;

  for (i=0; str[i]; i++) {
	switch (str[i]) {
		case '"': fputs("&quot;", f); break;
		case '<': fputs("&lt;", f); break;
		case '>': fputs("&gt;", f); break;
		case '&': fputs("&amp;", f); break;
		case 9:
		case 10:
		case 13: break;
		default: fputc(str[i], f); break;
	}
  }
}

struct varinf {
	int ofs;
	int size;
	int num;
	const char *name;
	const struct varinf *structvars;
	const char **flags;
};

char indent[11] = "\t\t\t\t\t\t\t\t\t\t";
#define INDENT (indent+10-indentation)

uint32_t getvalue(uint8_t *data, int size)
{
  uint32_t value = *(uint32_t*)(data);
  value &= 0xffffffff >> (32-8*size);

  return value;
}


int dumpflags(FILE *xml, uint8_t *data, const char **flags, int size, int indentation)
{
  uint32_t value;
  int i, state;
  const char *name, defname[32];

  if (size > 4) die("Can't dump flags with %d bytes", size);

  value = getvalue(data, size);

  for (i=0; flags[i]; i++) {
	name = flags[i];
	if (!strlen(name)) {
		sprintf(defname, "bit%d", i);
		name = defname;
	}
	state = (value & (1 << i)) >> i;
	if (strlen(flags[i]) || state)
		fprintf(xml, "%s<bit name=\"%s\">%d</bit>\n", INDENT, name, state);
  }
  return size;
}

int dodump(FILE *xml, uint8_t *data, struct varinf *vars, int indentation)
{
  int dumped, totaldumped, i, j, ofs;
  const char *name, *fname, defname[32], arrname[32];
fflush(xml);
  totaldumped = 0;


  for (i=0; vars[i].ofs != -1; i++) {
	fname = vars[i].name;
	if (!strlen(fname)) {
		sprintf(defname, "field_%X", vars[i].ofs);
		fname=defname;
	}
	ofs = vars[i].ofs;
	for (j=0; j<vars[i].num; j++) {
		name = fname;
		if (vars[i].num > 1) {
			sprintf(arrname, "%s[%d]", fname, j);
			name = arrname;
		}

		if (vars[i].flags) {
			fprintf(xml, "%s<bitmask name=\"%s\" size=\"%d\">\n", INDENT, name, vars[i].size);
			dumped = dumpflags(xml, data+ofs, vars[i].flags, vars[i].size, indentation+1);
			fprintf(xml, "%s</bitmask>\n", INDENT);
		} else if (vars[i].structvars) {
			fprintf(xml, "%s<structure name=\"%s\">\n", INDENT, name);
			dumped = dodump(xml, data+ofs, vars[i].structvars, indentation+1);
			fprintf(xml, "%s</structure>\n", INDENT);
		} else if (strlen(vars[i].name)) {
			fprintf(xml, "%s<variable name=\"%s\" size=\"%X\">%d</variable>\n", INDENT, name, vars[i].size, getvalue(data+ofs, vars[i].size));
			dumped = vars[i].size;
		} else {
			uint32_t value = getvalue(data+ofs, vars[i].size);
			if (value)
				fprintf(xml, "%s<unknown name=\"%s\" size=\"%X\">%d</unknown>\n", INDENT, name, vars[i].size, value);
			dumped = vars[i].size;
		}
		ofs += dumped;
		totaldumped += dumped;
	}
  }
  return totaldumped;
}

int dumpvars(FILE *xml, uint8_t *data, struct varinf *vars)
{
  return dodump(xml, data, vars, 1);
}

const char *spriteflags[] = {
	"hassprites",
	NULL
};

struct varinf vehsprites[] = {
	{ 0x00, 1, 1, "numdir" },
	{ 0x01, 1, 1, "" },
	{ 0x02, 1, 1, "" },
	{ 0x03, 1, 1, "" },
	{ 0x04, 1, 1, "" },
	{ 0x05, 1, 1, "" },
	{ 0x06, 1, 1, "" },
	{ 0x07, 1, 1, "flags", NULL, spriteflags },
	{ 0x08, 1, 1, "" },
	{ 0x09, 1, 1, "" },
	{ 0x0A, 1, 1, "" },
	{ 0x0B, 1, 1, "" },
	{ 0x0C, 1, 1, "" },
	{ 0x0D, 1, 1, "" },
	{ 0x0E, 1, 1, "spritenum" },
	{ 0x12, 4, 1, "" },
	{ 0x16, 4, 1, "" },
	{ 0x1A, 4, 1, "" },
	{ -1 },
};

struct varinf vehunk[] = {
	{ 0x00, 1, 1, "" },
	{ 0x01, 1, 1, "" },
	{ 0x02, 1, 1, "" },
	{ 0x03, 1, 1, "" },
	{ 0x04, 1, 1, "spriteind" },
	{ 0x05, 1, 1, "" },
	{ -1 },
};

const char *vehflags[] = {
	"", "", "", "",
	"", "", "rackrail", "",
	"", "", "", "",
	"dualhead", "", "refittable", "noannounce",
	NULL
};

struct varinf vehvars[] = {
	{ 0x000, 1, 2, "" },
	{ 0x002, 1, 1, "class" },
	{ 0x003, 1, 1, "" },
	{ 0x004, 1, 1, "" },
	{ 0x005, 1, 1, "" },
	{ 0x006, 1, 1, "nummods" },
	{ 0x007, 1, 1, "costind" },
	{ 0x008, 2, 1, "costfact" },
	{ 0x00A, 1, 1, "" },
	{ 0x00B, 1, 1, "runcostind" },
	{ 0x00C, 2, 1, "runcostfact" },
	{ 0x00E, 1, 1, "" },
	{ 0x00F, 1, 1, "numcompat" },
	{ 0x010, 1,20, "" },
	{ 0x024, 0, 4, "", vehunk },
	{ 0x03C, 0, 4, "sprites", vehsprites },
	{ 0x0B4, 1,36, "" },
	{ 0x0D8, 2, 1, "power" },
	{ 0x0DA, 2, 1, "speed" },
	{ 0x0DC, 2, 1, "rackspeed" },
	{ 0x0DE, 2, 1, "weight" },
	{ 0x0E0, 2, 1, "flags", NULL, vehflags },
	{ 0x0E2, 1,44, "" },
	{ 0x10E, 1, 1, "" },
	{ 0x10F, 1, 1, "visfxtype" },
	{ 0x110, 1, 1, "" },
	{ 0x111, 1, 1, "" },
	{ 0x112, 1, 1, "wakefxtype" },
	{ 0x113, 1, 1, "" },
	{ 0x114, 2, 1, "designed" },
	{ 0x116, 2, 1, "obsolete" },
	{ 0x118, 1, 1, "" },
	{ 0x119, 1, 1, "startsndtype" },
	{ 0x11A, 1,64, "" },
	{ 0x15A, 1, 1, "numsnd" },
	{ 0x15B, 1, 3, "" },
	{ -1 },
};

int dumpveh(FILE *xml, uint8_t *data)
{
  int dumped;

  dumped = dumpvars(xml, data, vehvars);

  return dumped;
}

int main(int argc, char **argv)
{
  FILE *dmp, *xml;
  char *dmpname, *xmlname, *pngname;
  struct stat st;

  uint8_t *data, *end;
  int class;
  uint32_t subclass;

  if (argc < 2) die("Give filename as parameter");

  dmpname = argv[1];

  dmp = fopen(dmpname, "rb");
  if (!dmp) die("Can't read %s: %s", dmpname, strerror(errno));

  if (fstat(fileno(dmp), &st)) die("Can't stat %s: %s", dmpname, strerror(errno));

#if defined(__BORLANDC__) && __BORLANDC__ < 0x500
  if (st.st_size > 65000) st.st_size = 65000;
#endif

  data = allocate(st.st_size, "data");

  fread(data, 1, st.st_size, dmp);
  if (fclose(dmp)) die("Can't close %s: %s", dmpname, strerror(errno));

  xmlname = allocate(strlen(dmpname)+1, "xmlname");

  pngname = allocate(strlen(dmpname)+1, "pngname");

  strcpy(xmlname, dmpname);
  strrchr(xmlname, '.')[0] = 0;
  strcat(xmlname, ".xml");

  strcpy(pngname, dmpname);
  strrchr(pngname, '.')[0] = 0;
  strcat(pngname, ".png");

  xml = fopen(xmlname, "wt");
  if (!xml) die("Can't write %s: %s", xmlname, strerror(errno));

  end = data+st.st_size;

  class = data[0];
  subclass = (*(uint32_t*)(data+1))&0xffffff;

  fputs("<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>\n", xml);
  fprintf(xml, "<object class=\"%02X\" subclass=\"%06X\" name=\"%.8s\">\n",
	class, subclass, data+4);
  data += 16;

  switch (class & 0x7f) {
	case 0x17:
		dumpveh(xml, data);
		break;
	default:
		die("Can't translate class %02X yet", class);
  }

  fputs("</object>\n", xml);
  fclose(xml);

  return 0;
}
