#include "common.h"

void rledecode(uint8_t *data, int len, FILE *dmp)
{
  while (len>0) {
	int8_t rle;
	int run;

	rle = *(int8_t*) data++; len--;
	run = abs(rle)+1;

	if (rle<0) {
		int i;
		uint8_t byte = *data++; len--;

		for (i=0; i<run; i++) fputc(byte, dmp);
	} else {
		fwrite(data, 1, run, dmp);
		len -= run;
		data += run;
	}
  }
}

int main(int argc, char **argv)
{
  FILE *obj, *dmp;
  char *objname, *dmpname;
  struct stat st;

  uint8_t *data, *end;

  if (argc < 2) die("Give filename as parameter");

  objname = argv[1];

  obj = fopen(objname, "rb");
  if (!obj) die("Can't read %s: %s", objname, strerror(errno));

  if (fstat(fileno(obj), &st)) die("Can't stat %s: %s", objname, strerror(errno));

#if defined(__BORLANDC__) && __BORLANDC__ < 0x500
  if (st.st_size > 65000) st.st_size = 65000;
#endif

  data = allocate(st.st_size, "data");

  fread(data, 1, st.st_size, obj);
  if (fclose(obj)) die("Can't close %s: %s", objname, strerror(errno));

  dmpname = allocate(strlen(objname)+1, "dmpname");

  strcpy(dmpname, objname);
  strrchr(dmpname, '.')[0] = 0;
  strcat(dmpname, ".dmp");

  dmp = fopen(dmpname, "wb");
  if (!dmp) die("Can't write %s: %s", dmpname, strerror(errno));

  end = data+st.st_size;
  fwrite(data, 1, 16, dmp);	// flags, filename, checksum
  data += 16;

  while (data < end) {
	int code;
	uint32_t len;

	code = *data; data++;
	len = *(uint32_t*) data; data+=4;

	switch(code) {
		case 0:
			fwrite(data, 1, len, dmp);
			break;
		case 1:
			rledecode(data, len, dmp);
			break;
		default:
			die("Can't decode chunk type %d", code);
	}
	data += len;
  }
  if (fclose(dmp)) die("Can't close %s: %s", dmpname, strerror(errno));

  return 0;
}
